(*****************************************************
Listes doublement chanes avec modification en place.
Double-linked lists with in-place modification.

D. Monniaux

$Id: dlist.ml,v 1.1.1.1 1998/06/23 22:35:07 monniaux Exp $
*****************************************************)

type 'a t =
  {
    mutable contents: 'a;
    mutable next:     'a t;
    mutable previous: 'a t
  };;

(* Construction *)
let singleton contents =
  let rec node =
    {
      next = node;
      previous = node;
      contents = contents
    }
  in node;;

let insert_after contents point =
  let node =
    {
      next     = point.next;
      previous = point;
      contents = contents 
    }
  in
    point.next.previous <- node;
    point.next          <- node;
    node;;

let insert_before contents point =
  let node =
    {
      previous = point.previous;
      next     = point;
      contents = contents 
    }
  in
    point.previous.next <- node;
    point.previous      <- node;
    node;;

(* Accs *)
let next x = x.next;;
let previous x = x.previous;;
let contents x = x.contents;;
let replace node newv = node.contents <- newv;;
let delete node =
  node.next.previous <- node.previous;
  node.previous.next <- node.next;;

(* Insertion de listes *)
let paste_after sublist point =
  sublist.previous.next <- point.next;
  point.next.previous <- sublist.previous;
  sublist.previous <- point;
  point.next <- sublist;;

let paste_before sublist point =
  point.previous.next <- sublist;
  sublist.previous.next <- point;
  let x = point.previous in
  ( point.previous <- sublist.previous;
    sublist.previous <- x );;
  
  
(**************************************************
Fonctions analogues  celles du module List.
Functions analogous to the ones in the List module.
**************************************************)

(* Iterators *)
let iter =
  let rec liter f where limit =
    f (where.contents);
    let follow = where.next
    in
      if (follow != limit)
      then liter f follow limit
  in fun f x -> liter f x x;;

let fold_left =
  let rec liter f where limit x =
    let follow = where.next and value = f x where.contents
    in
      if (follow != limit)
      then liter f follow limit value
      else value
  in fun f init x -> liter f x x init;;

let fold_right =
  let rec liter f where limit x =
    let follow = where.next
    in
      f where.contents
        ( if (follow != limit)
          then liter f follow limit x
          else x )
  in fun f x init -> liter f x x init;;

let map =
  let rec liter = fun f limit where point ->
    if (limit != where)
    then ( insert_before (f where.contents) point;
           liter f limit where.next point )
  in fun f list ->
    let org = singleton (f list.contents)
    in  liter f list list.next org; org;;

(* Iterators *)
let p_iter =
  let rec liter f where limit =
    f where;
    let follow = where.next
    in
      if (follow != limit)
      then liter f follow limit

  in fun f x -> liter f x x;;

let p_fold_left =
  let rec liter f where limit x =
    let follow = where.next and value = f x where
    in
      if (follow != limit)
      then liter f follow limit value
      else value
  in fun f init x -> liter f x x init;;

let p_fold_right =
  let rec liter f where limit x =
    let follow = where.next
    in
      f where
        ( if (follow != limit)
          then liter f follow limit x
          else x )
  in fun f x init -> liter f x x init;;

(* Divers *)

let length list = fold_right (function _ -> succ) list 0;;

let rec nth list = function
    0          -> list.contents
  | n when n<0 -> nth list.previous (succ n)
  | n          -> nth list.next (pred n);;

let rev =
  let rec liter = fun limit where point ->
    if (limit != where)
    then ( insert_after where.contents point;
           liter limit where.next point )
  in fun list ->
    let org = singleton list.contents
    in  liter list list.next org; org.next;;

(*
let toto = singleton 1;;
insert_after 2 toto;
insert_before 3 toto;
iter print_int toto;;

let tata = singleton 4;;
insert_after 5 tata;
insert_before 6 tata;
iter print_int tata;;

paste_after tata toto;;
iter print_int toto;;
*)

