(*******************************************************************
Librairie de graphes (oprations lmentaires)

- matrices d'adjacences

$Id: graph.ml,v 1.1.1.1 1998/06/23 22:35:07 monniaux Exp $
*******************************************************************)

module type GRAPH =
sig
  type t
  val create: int->t

  val add_edge: t->int->int->unit
  val rem_edge: t->int->int->unit
  val test_edge: t->int->int->bool

  val add_vertex: t->int->unit
  val rem_vertex: t->int->unit
  val test_vertex: t->int->bool

  val vertex_range: t->int
end;;

module Matrix : GRAPH =
struct
  type t = (bool array) * (bool array array);;
  let create n = (Array.create n false), (Array.create_matrix n n false);;

  let add_vertex = function v,e -> fun a ->
    v.(a) <-true;;
  let rem_vertex = function v,e -> fun a ->
    for i = 0 to pred (Array.length v)
    do
      e.(i).(a)<-false; e.(a).(i)<-false
    done;
    v.(a) <-false;;
  let test_vertex = function v,e -> fun a ->
    v.(a);;

  let vertex_range = function v,e -> Array.length v;;

  let check v a b =
    if a=b then raise (Invalid_argument "circular edge");
    if not (v.(a) & v.(b)) then raise (Invalid_argument "bad vertex");;

  let add_edge = function v,e -> fun a b ->
    check v a b;
    e.(a).(b)<-true; e.(b).(a)<-true;;
  let rem_edge = function v,e -> fun a b ->
    check v a b;
    e.(a).(b)<-false; e.(b).(a)<-false;;
  let test_edge = function v,e -> fun a b ->
    check v a b;
    e.(a).(b)
end;;





