(*
$Id: lexer.mll,v 1.1.1.1 1998/06/23 22:35:08 monniaux Exp $
*)

{
open Parser;;

let line_offsets = ref [0,1];;
let cur_line = ref 2;;

let location_of_pos pos =
  let rec search =
    function
      (posl, ln)::_ when posl<=pos -> ln, (pos-posl)
    | _::t -> search t
  (* pattern matching non exhaustive *)
  in search !line_offsets;;
(* Donne ((l1, c1),(l2, c2)) avec lx et cx le numro de ligne
   (commenant  1) et de caractre dans la ligne (commenant  0)
   des premier et dernier caractres du token courant (qui a provoqu
   une erreur *)

let error_extent lexbuf =
   (location_of_pos (Lexing.lexeme_start lexbuf)),
   (location_of_pos (pred (Lexing.lexeme_end lexbuf)));;

let start_of_comment = ref 0;;
}

rule token = parse
   [' ' '\t' '\r']+ { token lexbuf }
 | '\n' {
          line_offsets :=
            ((Lexing.lexeme_end lexbuf), !cur_line):: !line_offsets;
          incr cur_line;
          token lexbuf }
 | "(*" { start_of_comment := Lexing.lexeme_start lexbuf;
          comment lexbuf ; token lexbuf }
 | "and"       { AND }
 | "array"     { ARRAY }
 | "begin"     { BEGINBLOCK }
 | "case"      { CASE }
 | "declare"   { DECLARE }
 | "div"       { DIV }
 | "do"        { DO }
 | "else"      { ELSE }
 | "end"       { END }
 | "for"       { FOR }
 | "from"      { FROM }
 | "if"        { IF }
 | "is"        { IS }
 | "mod"       { MOD }
 | "not"       { NOT }
 | "of"        { OF }
 | "or"        { OR }
 | "procedure" { PROCEDURE }
 | "record"    { RECORD }
 | "ref"       { REF }
 | "new"       { NEW }
 | "then"      { THEN }
 | "to"        { TO }
 | "type"      { TYPE }
 | "while"     { WHILE }
 | ['0'-'9']+      { Integer(int_of_string (Lexing.lexeme lexbuf)) }
 | (['0'-'9']+ ['e' 'E'] ['-' '+']? ['0'-'9']+) |
   (['0'-'9']+ '.' ['0'-'9']* (['e' 'E'] ['-' '+']? ['0'-'9']+)?)
               { Float(float_of_string (Lexing.lexeme lexbuf)) }
 | ['A'-'Z' 'a'-'z'] ('_'? ['A'-'Z' 'a'-'z' '0'-'9'])*
               { Identifier(Lexing.lexeme lexbuf) }
 | ":="        	{ ASSIGN }
 | "<"		{ INFERIOR }
 | ">"		{ SUPERIOR }
 | "<="		{ INFERIOR_EQUAL }
 | ">="		{ SUPERIOR_EQUAL }
 | "<>"		{ UNEQUAL }
 | "="		{ EQUAL }
 | "=="		{ IDENTICAL }
 | "!="		{ UNIDENTICAL }
 | "+"		{ ADD }
 | "-"		{ SUB }
 | "*"		{ MUL }
 | "/"		{ DIV2 }
 | "("		{ LPAREN }
 | ")"		{ RPAREN }
 | "["		{ LBRACKET }
 | "]"		{ RBRACKET }
 | ";"		{ SEMICOLON }
 | ","		{ COMMA }
 | ":"		{ COLON }
 | "."		{ DOT }
 | "^"		{ WEDGE }
 | eof  	{ Eof }

 | [^ 'A'-'Z' 'a'-'z' ':' '=' '<' '>' '=' '+' '-' '*' '/' '(' ')'
        '[' ']' ';' ',' ':' '.' '^' ' ' '\t' '\n' ]+
		{
                  Error.syntax_error "lexical error" (error_extent lexbuf);
                  token lexbuf
                }

and comment = parse
    "(*" { comment lexbuf ; comment lexbuf }
  | "*)" { () }
  | '\n' {
          line_offsets :=
            ((Lexing.lexeme_end lexbuf), !cur_line):: !line_offsets;
          incr cur_line;
          comment lexbuf }
  | _    { comment lexbuf }
  | eof  { Error.syntax_error "unterminated comment"
             ((location_of_pos !start_of_comment),
              (location_of_pos (pred (Lexing.lexeme_start lexbuf)) ));
           raise Exit }
