(*
$Id: main.ml,v 1.1.1.1 1998/06/23 22:35:08 monniaux Exp $
*)

open Syntax;;

open Typeintermed;;

let filename = ref "test4.lax";;

let rec genere = function 
  | [] -> failwith "je ne sait plus programmer..."
  | [(ut, mem, id, nb)] -> (try Sparcgen.generation ut mem id nb
    with
        Match_failure(f,a,b) -> Error.unimplemented "code generation" f a b
      | Failure(x) -> Error.failure "code generation" x
    )
  | x::l -> genere [x] ; genere l;;


  Arg.parse [
"-fpack-cc", (Arg.Set Options.shoot_subcc_0), "\t\t\tturn on condition code folding";
"-fno-pack-cc", (Arg.Clear Options.shoot_subcc_0), "\t\t\tturn off condition code folding";

"-ffollow-branches", (Arg.Set Options.follow_on_branch), "\t\tturn on branch following";
"-fno-follow-branches", (Arg.Clear Options.follow_on_branch), "\t\tturn off branch following";

"-fremove-unneeded-branches", (Arg.Set Options.remove_unneeded_branches), "\tturn on removal of unneeded branches, first pass";
"-fno-remove-unneeded-branches", (Arg.Clear Options.remove_unneeded_branches), "turn off removal of unneeded branches, first pass";

"-fremove-unneeded-branches2", (Arg.Set Options.remove_unneeded_branches2), "\tturn on removal of unneeded branches, second pass";
"-fno-remove-unneeded-branches2", (Arg.Clear Options.remove_unneeded_branches2), "turn off removal of unneeded branches, second pass";

"-fextended-remove-branches", (Arg.Set Options.extended_remove_branches), "\tturn on extended removal of unneeded branches";
"-fno-extended-remove-branches", (Arg.Clear Options.extended_remove_branches), "turn off extended removal of unneeded branches";

"-fremove-unneeded-nops", (Arg.Set Options.remove_unneeded_nops), "\tturn on removal of unneeded NOPs";
"-fno-remove-unneeded-nops", (Arg.Clear Options.remove_unneeded_nops), "\tturn off removal of unneeded NOPs";

"-fremove-dead-code", (Arg.Set Options.remove_dead_code), "\t\tturn on removal of dead code";
"-fno-remove-dead-code", (Arg.Clear Options.remove_dead_code), "\tturn off removal of dead code";

"-ffill-delay-slot", (Arg.Set Options.fill_delay_slot), "\t\tturn on filling of delay slots";
"-fno-fill-delay-slot", (Arg.Clear Options.fill_delay_slot), "\t\tturn off filling of delay slots";

"-fsimplify", (Arg.Set Options.simplify), "\t\tturn on expression simplification";
"-fno-simplify", (Arg.Clear Options.simplify), "\t\tturn off expression simplification";

"-fno-peephole", (Arg.Unit (function () -> 
    Options.shoot_subcc_0 :=false;
    Options.follow_on_branch :=false;
    Options.remove_unneeded_branches :=false;
    Options.remove_unneeded_branches2 :=false;
    Options.remove_unneeded_nops :=false;
    Options.fill_delay_slot :=false;
    Options.remove_dead_code :=false;
      )), "\t\tturn off all peephole optimizations"
    ] (function x->filename:=x) "bad usage";;

let lexbuf = Lexing.from_channel (open_in !filename);;

(
try
(
try
( let parsed = Parser.bidule Lexer.token lexbuf
in (*output_string stderr "Parsing ok.\n"; *)
   try
genere (Memoire.second_transform (Intermed.first_transform parsed));
Sparc.output_fconsts ()

with Match_failure(f,a,b) -> Error.unimplemented "semantic analysis" f a b
  | Failure(x) -> Error.failure "semantic analysis" x
)
with Match_failure(f,a,b) ->  Error.unimplemented "syntax analysis" f a b
| Parsing.Parse_error ->
    Error.syntax_error "unrecoverable parse error" (Lexer.error_extent lexbuf)
| Failure("empty token") ->
    let (ofs, ln)::_ = !Lexer.line_offsets in
                (* pattern matching non exhaustive *)
     Printf.fprintf stderr "Lexical error at line %d, char %d\n"
      ln ((Lexing.lexeme_end lexbuf)-ofs)
| Failure(x) -> Error.failure "syntax analysis" x
)
with Exit -> (Printf.fprintf stderr "Compilation aborted.\n"; exit 1) )
 ;;
 







