(*********************************************************
Gnration de code Sparc
version en sortie d'un allocateur de variables

$Id: sparcgen2.ml,v 1.1.1.1 1998/06/23 22:35:09 monniaux Exp $
*********************************************************)

open Sparc;;

open Typeintermed2;;

let nb_hardware_registers = 20
and nb_hardware_fregisters = 15
and parameters_reg = 24
and input_registers = 8
and parameters_offset = 68
and save_old_frame_offset = -8
and announce_integer_stack = 100000
and announce_float_stack = 200000
and bottom_of_frame_size = 24*4
and scratch_register2 = 2
and scratch_register3 = 8
and scratch_register4 = 9
and fscratch_register = 30
and fscratch_register2 = 28
and ret_vregister=0
and ret_vfregister=0;;

let dummy_reg = -1;;

let i_stack = ref 0 and f_stack = ref 0;;

let phys_reg x = x+8
and phys_freg x = x*2;;

let load_from (vreg: virtual_reg) (where: instruction) =
  if vreg<nb_hardware_registers then
    phys_reg vreg
  else
    ( i_stack:=max !i_stack (vreg-nb_hardware_registers);
      Dlist.insert_before (Opcode(Load(Word, frame_pointer,
        (Imm13 (announce_integer_stack+vreg-nb_hardware_registers)),
        scratch_register))) where;
      scratch_register )

and load_from2 (vreg: virtual_reg) (where: instruction) =
  if vreg<nb_hardware_registers then
    phys_reg vreg
  else
    ( i_stack:=max !i_stack (vreg-nb_hardware_registers);
      Dlist.insert_before (Opcode(Load(Word, frame_pointer,
        (Imm13 (announce_integer_stack+vreg-nb_hardware_registers)),
        scratch_register2))) where;
      scratch_register2 )

and load_from3 (vreg: virtual_reg) (where: instruction) =
  if vreg<nb_hardware_registers then
    phys_reg vreg
  else
    ( i_stack:=max !i_stack (vreg-nb_hardware_registers);
      Dlist.insert_before (Opcode(Load(Word, frame_pointer,
        (Imm13 (announce_integer_stack+vreg-nb_hardware_registers)),
        scratch_register3))) where;
      scratch_register3 )

and save_to_reg (vreg: virtual_reg) =
  if vreg<nb_hardware_registers then
    phys_reg vreg
  else
    scratch_register

and save_to (vreg: virtual_reg) (where: instruction) =
  ( if vreg>=nb_hardware_registers then
      ( i_stack:=max !i_stack (vreg-nb_hardware_registers);
        Dlist.insert_before
          (Opcode(Store(Word, scratch_register, frame_pointer,
          (Imm13 (announce_integer_stack+vreg-nb_hardware_registers)) )))
        where; ()
      )
  ); where
;;

let fload_from (vreg: virtual_reg) (where: instruction) =
  if vreg<nb_hardware_fregisters then
    phys_freg vreg
  else
    ( f_stack:=max !f_stack (vreg-nb_hardware_fregisters);
      Dlist.insert_before (Opcode(F_Load(Double, frame_pointer,
        (Imm13 (announce_integer_stack+vreg-nb_hardware_registers)),
        fscratch_register))) where;
      fscratch_register )

and fload_from2 (vreg: virtual_reg) (where: instruction) =
  if vreg<nb_hardware_fregisters then
    phys_freg vreg
  else
    ( f_stack:=max !f_stack (vreg-nb_hardware_fregisters);
      Dlist.insert_before (Opcode(F_Load(Double, frame_pointer,
        (Imm13 (announce_integer_stack+vreg-nb_hardware_registers)),
        fscratch_register2))) where;
      fscratch_register2 )

and fsave_to_reg (vreg: virtual_reg) =
  if vreg<nb_hardware_registers then
    phys_freg vreg
  else
    fscratch_register

and fsave_to (vreg: virtual_reg) (where: instruction) =
  ( f_stack:=max !f_stack (vreg-nb_hardware_fregisters);
    if vreg>=nb_hardware_fregisters then
      ( Dlist.insert_before
        (Opcode(F_Store(Double, fscratch_register, frame_pointer,
        (Imm13 (announce_integer_stack+vreg-nb_hardware_registers)) ))) where;
        () )
  ); where
;;

let check_regs r1 r2 = if r1<>r2 then failwith "incompatible register mapping";;

let rec create_args (index: int) (block1: instruction) =
  function
    [] -> ()
  | h::tail ->
      let rs=( match h with
          Pexpr((vp,p)) -> Dlist.paste_before (ptr_eval (vp,p)) block1;
                           load_from vp block1
        | Iexpr((vi,i)) -> Dlist.paste_before (int_eval (vi,i)) block1;
                           load_from vi block1
      ) in ( if index>=6 then
               ( Dlist.insert_before
                 (Opcode(Store(Word, rs, stack_pointer,
                 (Imm13 (parameters_offset+4*index))))) block1; () )
             else
               (if rs<>parameters_reg+index
                then failwith "bad register for parameter passing")
           );
        create_args (succ index) block1 tail

and create_call (ptr: pexpr) (args: expr list) =
  (let nb_args = List.length args
  and block1 = Dlist.singleton (Opcode(Nop))
  in create_args 0 block1 args;
     ( match ptr with
         _,(AsmLabel l) ->
           Dlist.insert_before (Opcode(Call l)) block1
       | rp,_ -> Dlist.paste_before (ptr_eval ptr) block1;
                 let r1=load_from rp block1
                 in Dlist.insert_before (Opcode(Callr (r1))) block1
     ); Dlist.insert_before (Opcode Nop) block1;
          let shift = ((nb_args-6)*4+7) land -8
          in if shift>0
             then ( Dlist.insert_before (Opcode(
  Triadic(Add, false, stack_pointer, (Imm13 shift), stack_pointer))) block1;
                    (Dlist.insert_before (Opcode(
  Triadic(Sub, false, stack_pointer, (Imm13 shift), stack_pointer))) block1)
                  )
             else block1
)

and branch_if (e : bexpr) (target: Sparc.label) =
  match e with
    BNot(BNot(x)) -> branch_if x target

  | True -> create_branch Always target
  | BNot(True) -> branch_if False target

  | False -> (Dlist.singleton (Opcode(Nop)))
  | BNot(False) -> branch_if True target

  | BOr(x,y) -> let block1 = branch_if x target
                and block2 = branch_if y target
                in  Dlist.paste_before block2 block1;
                    block1

  | BNot(BOr(x,y)) -> branch_if (BAnd((BNot(x)),(BNot(y)))) target

  | BAnd(x,y) -> let label = make_label()
                 in  let block1 = branch_if (BNot x) label
                     and block2 = branch_if y target
                     in  Dlist.paste_before block2 block1;
                         Dlist.insert_before (Label(label)) block1;
                         block1

  | BNot(BAnd(x,y)) -> branch_if (BOr((BNot(x)),(BNot(y)))) target

  | BEgal(a, b) -> let label1 = make_label() and label2 = make_label()
                   in  let block1 = branch_if a label1
                       and block2 = branch_if (BNot b) target
                       and block3 = branch_if b target
                       in  Dlist.paste_before block2 block1;
                           Dlist.paste_before (create_branch Always label2) block1;
                           Dlist.insert_before (Label label1) block1;
                           Dlist.paste_before block3 block1;
                           Dlist.insert_before (Label label2) block1;
                           block1

  | BNot(BEgal(a, b)) -> branch_if (BEgal((BNot a),b)) target

  | IInf(a, b) -> int_compare Less target a b
  | BNot(IInf(a, b)) -> int_compare GreaterOrEqual target a b

  | ISup(a, b) -> int_compare Greater target a b
  | BNot(ISup(a, b)) -> int_compare LessOrEqual target a b

  | IEgal(a, b) -> int_compare Equal target a b 
  | BNot(IEgal(a, b)) -> int_compare NotEqual target a b 

  | PEgal(a, b) -> ptr_compare Equal target a b 
  | BNot(PEgal(a, b)) -> ptr_compare NotEqual target a b 

  | FInf(a, b) -> float_compare F_Less target a b
  | BNot(FInf(a, b)) -> float_compare F_GreaterOrEqual target a b

  | FSup(a, b) -> float_compare F_Greater target a b
  | BNot(FSup(a, b)) -> float_compare F_LessOrEqual target a b

  | FEgal(a, b) -> float_compare F_Equal target a b 
  | BNot(FEgal(a, b)) -> float_compare F_NotEqual target a b 

  | Bseq(a, b) -> void_eval a; branch_if b target
  | BNot(Bseq(a, b)) -> branch_if (Bseq(a, (BNot b))) target

  | B_ITE(a, b, c) -> let label = make_label() and label2=make_label()
                     in let block1 = branch_if (BNot a) label
                        and block2 = branch_if b target
                        and block3 = branch_if c target
                        in  Dlist.paste_before block2 block1;
                            Dlist.paste_before (create_branch Always label2) block1;
                            Dlist.insert_before (Label label) block1;
                            Dlist.paste_before block3 block1;
                            Dlist.insert_before (Label label2) block1;
                            block1

  | BNot(B_ITE(a, b, c)) -> branch_if (B_ITE(a, (BNot b), (BNot c))) target

  | Binvseq(a, b) -> let label = make_label()
                     in let block1 = branch_if (BNot a) label
                        and block2 = void_eval b
                        and block3 = void_eval b
                        in  Dlist.paste_before block2 block1;
                            Dlist.paste_before (create_branch Always target) block1;
                            Dlist.insert_before (Label label) block1;
                            Dlist.paste_before block3 block1;
                            block1

  | BNot(Binvseq(a, b)) -> branch_if (Binvseq((BNot a), b)) target

  (*| Bprocedure(ptr, arg) ->
      branch_if (BNot(IEgal((Iprocedure(ptr, arg)),(IConst 0)))) target

  | BNot(Bprocedure(ptr, arg)) ->
      branch_if (IEgal((Iprocedure(ptr, arg)),(IConst 0))) target*)

  | B_peek(ptr) -> branch_if (BNot(IEgal((dummy_reg,(I_peek ptr)),(dummy_reg,(IConst 0))))) target
  | BNot(B_peek(ptr)) -> branch_if (IEgal((dummy_reg,(I_peek ptr)),(dummy_reg,(IConst 0)))) target
  | B_reg(reg) -> branch_if (BNot(IEgal((reg,I_reg),(dummy_reg,(IConst 0))))) target
  | BNot(B_reg reg) -> branch_if (IEgal((reg,I_reg),(dummy_reg,(IConst 0)))) target

and int_compare (typ : condition) (target: label)
  ((ar,a) as ap: iexpr) ((br,b) as bp: iexpr) =
      match b with
        IConst(c) when c>= -4096 & c<4096 ->
          let block1 = int_eval ap
          in let r1=load_from ar block1
          in Dlist.insert_before
                (Opcode(Triadic(Sub, true, r1, Imm13(c), 0))) block1;
              Dlist.paste_before (create_branch typ target) block1;
              block1
      | _ -> let block2 = int_eval bp
             and block1 = int_eval ap
             in let r1=load_from ar block1
             in let r2=load_from2 br block2
             in  Dlist.paste_before block2 block1;
                 Dlist.insert_before
                   (Opcode(Triadic(Sub, true, r1, (Reg r2), 0)))
                 block1;
            Dlist.paste_before (create_branch typ target) block1;
            block1

and ptr_compare (typ : condition) (target: label)
  ((ar, a) as ap: pexpr) ((br, b) as bp: pexpr) =
           let block2 = ptr_eval bp
           and block1 = ptr_eval ap
           in let r1=load_from ar block1
           in let r2=load_from2 br block2
           in  Dlist.paste_before block2 block1;
                     Dlist.insert_before
                       (Opcode(Triadic(Sub, true, r1, (Reg r2), 0)))
                     block1;
            Dlist.paste_before (create_branch typ target) block1;
            block1

and float_compare (typ : f_condition) (target: label)
    ((ar, a) as ap: fexpr) ((br, b) as bp: fexpr) =
  let block2 = float_eval bp
  and block1 = float_eval ap
  in let r1=fload_from ar block1
  in let r2=fload_from2 br block2
  in  Dlist.paste_before block2 block1;
          Dlist.insert_before (Opcode(F_Cmp(r1 , r2))) block1;
          Dlist.insert_before (Opcode(Nop)) block1;
          Dlist.paste_before (create_fbranch typ target) block1;
          block1

and int_eval ((r,ex): iexpr) =
  match ex with
    IConst(0) -> (Dlist.singleton (Opcode Nop))
  | IConst(x) -> save_to r (create_load_const x (save_to_reg r))

  | I_peek((_,PAdd((_,FramePtr), (_,IConst(ofs))))) ->
      save_to r (Dlist.singleton(Opcode(Load(Word, frame_pointer, Imm13(ofs),
        (save_to_reg r) ))))

  | I_peek((rx,x) as rp) ->
      let block1 = ptr_eval rp
      in let r1=load_from rx block1
      in  Dlist.insert_before
            (Opcode(Load(Word, r1, Reg(0), (save_to_reg r)))) block1;
          save_to r block1

  | INeg((rx,x) as rp) ->
      let block1 = int_eval rp
      in let r1=load_from rx block1
      in  Dlist.insert_before
            (Opcode(Triadic(Sub, false, 0, Reg(r1), (save_to_reg r)))) block1;
          save_to r block1

  | IAdd(x, y) -> create_triadic Add_op r x y
  | ISub(x, y) -> create_triadic Sub_op r x y
  | IMul(x, y) -> create_triadic Mul_op r x y
  | IDiv(x, y) -> create_triadic Div_op r x y
  | IMod(x, y) -> create_triadic Mod_op r x y

  | I_ITE(t, ((rx, x) as xp), ((ry, y) as yp)) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1 = int_eval xp
          and block2 = int_eval yp
          and block0 = branch_if t label1
          in  check_regs r rx; check_regs r ry;
              Dlist.paste_before block2 block0;
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              block0

  | Iseq(e1, ((r2, e2) as p2)) ->
                        let block1 = void_eval e1
                        and block2 = int_eval p2
                        in  check_regs r r2;
                            Dlist.paste_before block2 block1;
                            block1

  | Iinvseq(((r1, e1) as p1), e2) ->
                       let block2 = void_eval e2
                       and block1 = int_eval p1
                       in  check_regs r r1;
                           Dlist.paste_before block2 block1;
                           block1

  | Iprocedure(proc, args) ->
      ( if r<>ret_vregister then failwith "bad return register" );
      create_call proc args
  | I_reg -> Dlist.singleton (Opcode Nop)

and ptr_eval ((r,ex) : pexpr) =
  match ex with
    Nil -> (Dlist.singleton (Opcode Nop))
  | AsmLabel(x) -> save_to r (create_load_label x (save_to_reg r))
  | FramePtr -> save_to r (Dlist.singleton(Opcode(Triadic(Add, false,
      frame_pointer, Reg(0), (save_to_reg r)))))
  | PAdd(x, y) -> create_ptriadic Add_op r x y

  | P_ITE(t, ((r1, x) as p1), ((r2, y) as p2)) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1 = ptr_eval p1
          and block2 = ptr_eval p2
          and block0 = branch_if t label1
          in  check_regs r r1; check_regs r r2;
              Dlist.paste_before block2 block0;
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              block0

  | Pseq(e1, ((r2, e2) as p2)) ->
                        let block1 = void_eval e1
                        and block2 = ptr_eval p2
                        in  check_regs r r2;
                            Dlist.paste_before block2 block1;
                            block1

  | Pinvseq(((r1, e1) as p1), e2) ->
                       let block2 = void_eval e2
                       and block1 = ptr_eval p1
                       in  check_regs r r1;
                           Dlist.paste_before block2 block1;
                           block1

  | P_peek((_,PAdd((_,FramePtr), (_, (IConst ofs))))) ->
      save_to r (Dlist.singleton(Opcode(Load(Word, frame_pointer,
        Imm13(ofs), (save_to_reg r)))))

  | P_peek((rx,x) as xp) ->
      let block1 = ptr_eval xp
      in let r1 = load_from rx block1
      in  Dlist.insert_before (Opcode(Load(Word, r1, Reg(0),
            (save_to_reg r)))) block1;
          save_to r block1

  | FAssignment(((rp, p) as pp), ((rv, v) as pv)) ->
                         let block2 = ptr_eval pp
                         and block1 = float_eval pv
                         in check_regs r rv;
                         let r1 = load_from rp block1
                         in let r2 = load_from2 rv block1
                         in Dlist.paste_before block2 block1;
          Dlist.insert_before (Opcode(F_Store(Double, r1, r2, Reg(0)))) block1;
                             block1
   
  | PAssignment(((rp, p) as pp), ((rv, v) as pv)) ->
             let block2 = ptr_eval pp
             and block1 = ptr_eval pv
                        in  check_regs r rv;
                        let r1 = load_from rp block1
                        in let r2 = load_from2 rv block1
                        in Dlist.paste_before block2 block1;
          Dlist.insert_before (Opcode(Store(Word, r1, r2, Reg(0)))) block1;
                             block1

  | IAssignment(((rp, p) as pp), ((rv, v) as pv)) ->
             let block2 = ptr_eval pp
             and block1 = int_eval pv
                        in  check_regs r rv;
                        let r1 = load_from rp block1
                        in let r2 = load_from2 rv block1
                        in Dlist.paste_before block2 block1;
          Dlist.insert_before (Opcode(Store(Word, r1, r2, Reg(0)))) block1;
                             block1

   | Alloca((rx, s) as pr) ->
       let block1 = int_eval pr
       in let r1 = load_from rx block1
       in Dlist.insert_before
  (Opcode(Triadic(Add, false, stack_pointer, Reg(0), scratch_register))) block1;
           Dlist.insert_before
  (Opcode(Triadic(Add, false, r1, Imm13(15), r1))) block1;
           Dlist.insert_before
  (Opcode(Triadic(Sparc.And, false, r1, Imm13(-8), r1))) block1;
           Dlist.insert_before
  (Opcode(Triadic(Sub, false, stack_pointer, Reg(r1), stack_pointer))) block1;
           Dlist.insert_before (Opcode(
 Store(Word, scratch_register, stack_pointer, Imm13(bottom_of_frame_size)))) block1;
           Dlist.insert_before
  (Opcode(Triadic(Add, false, stack_pointer, Imm13(8+bottom_of_frame_size),
    (save_to_reg r)))) block1;
           save_to r block1

  | Pprocedure(proc, args) ->
      ( if r<>ret_vregister then failwith "bad return register" );
      create_call proc args

  | BAssignment(ptr, expr) ->
      ptr_eval (r,(IAssignment(ptr, (bool_to_int expr))))
  
  | P_reg -> Dlist.singleton (Opcode Nop)

and bool_to_int expr = (dummy_reg,I_ITE(expr, (dummy_reg,(IConst 1)), (dummy_reg, (IConst 0))))

and create_triadic (typ : binary_op) (r: virtual_reg)
  ((rx, x) as px: iexpr) ((ry, y) as py: iexpr) =
  match y with
        IConst(c) when c>= -4096 & c<4096
          -> let block1 = int_eval px
             in let r1 = load_from rx block1
             in  ( Dlist.paste_before
                (create_triadic_op typ r1 (Imm13 c) (save_to_reg r)) block1;
                        save_to r block1
                      )

      | _ -> let block2 = int_eval py
             and block1 = int_eval px
             in let r1 = load_from rx block1
             in let r2 = load_from2 ry block1
             in  save_to r (create_triadic_op typ r1 (Reg r2) (save_to_reg r))

and create_ptriadic (typ : binary_op) (r: virtual_reg)
  ((rx, x) as px: pexpr) ((ry, y) as py: iexpr) =
  match y with
        IConst(c) when c>= -4096 & c<4096
          -> let block1 = ptr_eval px
             in let r1 = load_from rx block1
             in  ( Dlist.paste_before
               (create_triadic_op typ r1 (Imm13 c) (save_to_reg r)) block1;
                        save_to r block1
                      )

      | _ -> let block2 = int_eval py
             and block1 = ptr_eval px
             in let r1 = load_from rx block1
             in let r2 = load_from ry block1
             in  Dlist.paste_before block2 block1;
                 Dlist.paste_before (create_triadic_op typ r1 (Reg r2) (save_to_reg r)) block1;
                 save_to r block1

and float_eval ((r, ex): fexpr) =
  match ex with
    F_from_i((rx, x) as px) -> let block1 = int_eval px
      in let r1 = fload_from rx block1
      in fsave_to r (create_dtoi r1 (fsave_to_reg r))

  | FNeg((rx, x) as px) -> let block1 = float_eval px
      in let r1 = fload_from rx block1
      in Dlist.insert_before(Opcode(F_Diadic(F_Neg, r1, (fsave_to_reg r)))) block1;
         fsave_to r block1
  | FAdd(x, y) -> create_ftriadic F_Add r x y
  | FSub(x, y) -> create_ftriadic F_Sub r x y
  | FMul(x, y) -> create_ftriadic F_Mul r x y
  | FDiv(x, y) -> create_ftriadic F_Div r x y
  | F_ITE(t, ((rx, x) as px), ((ry, y) as py)) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1 = float_eval px
          and block2 = float_eval py
          and block0 = branch_if t label1
          in  check_regs r rx; check_regs r ry;
              Dlist.paste_before block2 block0;
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              block0

  | F_peek((_,AsmLabel(l))) ->
      let block1 =
     Dlist.singleton(Opcode(Sethi((Hi_l l), scratch_register)))
      in Dlist.insert_before
       (Opcode(F_Load(Double, scratch_register, (Lo_l l), (fsave_to_reg r)))) block1;
                   fsave_to r block1

  | F_peek((_,PAdd((_,FramePtr), (_,(IConst ofs))))) ->
      save_to r (Dlist.singleton(Opcode(F_Load(Double, frame_pointer,
        Imm13(ofs), (fsave_to_reg r)))))

  | F_peek((rp, p) as pp) -> let block1=ptr_eval pp
      in let r1 = fload_from rp block1
      in Dlist.insert_before (Opcode(F_Load(Double, r1, (Reg 0),
           (fsave_to_reg r)))) block1;
                   fsave_to r block1

  | FConst(x) -> float_eval (dummy_reg,(F_peek(dummy_reg, (AsmLabel (gen_fconst_label (hash_into float_consts x float_consts_index))))))

  | Fprocedure(proc, args) ->
    ( if r<>ret_vfregister then failwith "bad float return register" );
    create_call proc args

  | Fseq(e1, ((r2, e2) as p2)) ->
                        let block1 = void_eval e1
                        and block2 = float_eval p2
                        in  check_regs r r2;
                            Dlist.paste_before block2 block1;
                            block1

  | Finvseq(((r1, e1) as p1), e2) ->
                       let block2 = void_eval e2
                       and block1 = float_eval p1
                       in  check_regs r r1;
                           Dlist.paste_before block2 block1;
                           block1
  | F_reg -> Dlist.singleton (Opcode Nop)

and create_ftriadic (typ : f_triadic) (r: virtual_reg)
    ((rx, x) as px: fexpr) ((ry, y) as py: fexpr) =
  let block2 = float_eval py
  and block1 = float_eval px
  in let r1 = fload_from rx block1
  in let r2 = fload_from2 ry block1
  in fsave_to r (create_ftriadic_op typ r1 r2 (fsave_to_reg r))

and void_eval =
  function
  (*  V_from_p(IAssignment(PAdd(FramePtr, (IConst ofs)), v)) ->
      let block1, s1, sf1, r1 = int_eval v 0
      in
Dlist.insert_before (Opcode(Store(Word, r1, frame_pointer, (Imm13 ofs)))) block1;
          block1, s1, sf1

  | V_from_p(PAssignment(PAdd(FramePtr, (IConst ofs)), v)) ->
      let block1, s1, sf1, r1 = ptr_eval v 0
      in
Dlist.insert_before (Opcode(Store(Word, r1, frame_pointer, (Imm13 ofs)))) block1;
          block1, s1, sf1

  | V_from_p(FAssignment(PAdd(FramePtr, (IConst ofs)), v)) ->
      let block1, s1, sf1, r1 = float_eval v 0
      in
Dlist.insert_before(Opcode(F_Store(Double, r1, frame_pointer, (Imm13 ofs)))) block1;
          block1, s1, sf1
*)

  | V_from_i(iexpr) -> int_eval iexpr
  | V_from_p(pexpr) -> ptr_eval pexpr
  | V_from_f(fexpr) -> float_eval fexpr
  | V_from_b(bexpr) -> let label=make_label()
                       in  let block1 = branch_if bexpr label
                           in  Dlist.insert_before (Label label) block1;
                               block1

  | Vseq(e1, e2)    -> let block1 = void_eval e1
                       and block2 = void_eval e2
                       in  Dlist.paste_before block2 block1;
                           block1

  | V_ITE(t, x, y) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1 = void_eval x
          and block2 = void_eval y
          and block0 = branch_if t label1
          in  Dlist.paste_before block2 block0;
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              block0

  | V_IT(t, x) ->
      let label1 = make_label()
      in  let block1 = void_eval x
          and block0 = branch_if (BNot t) label1
          in  Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label1)) block0;
              block0

  | While(t, x) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1 = void_eval x
          and block2 = branch_if t label1
          and block0 = create_branch Always label2
          in  Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              Dlist.paste_before block2 block0;
              block0

  | Free -> (Dlist.singleton (Opcode
   (Load(Word, stack_pointer, Imm13(bottom_of_frame_size), stack_pointer))))

  | Erreur -> let block1 = Dlist.singleton (Opcode(Call "lax__bad_dimensions"))
              in  Dlist.insert_before (Opcode(Nop)) block1;
                  block1

  | NNop -> (Dlist.singleton (Opcode(Nop)))

  | CopyBlock(src, dst, len) -> create_copyblock src dst len Word
  | FCopyBlock(src, dst, len) -> create_copyblock src dst len DoubleWord

  | Vprocedure(proc, args) -> create_call proc args

and create_copyblock ((rsrc,src) as s) ((rdst,dst) as d) ((rlen,len) as l)
  typ =
      let block1 = ptr_eval s
      in  (Dlist.paste_before (ptr_eval d) block1;
          Dlist.paste_before (int_eval l) block1;
          let r1=load_from rsrc block1
          and r2=load_from2 rdst block1
          and r3=load_from3 rlen block1
          and label1 = make_label () and label2 = make_label()
          in Dlist.insert_before
               (Opcode(Triadic(Add, true, r3, Reg(0), 0))) block1;
             Dlist.paste_before (create_branch Less label2) block1;
             Dlist.insert_before (Label(label1)) block1;
         ( match typ with
             Word ->
               Dlist.insert_before (Opcode(Load(Word, r1, Reg(r3), scratch_register4))) block1;
               Dlist.insert_before (Opcode(Store(Word, scratch_register4, r2, Reg(r3)))) block1;
           | DoubleWord ->
               Dlist.insert_before (Opcode(F_Load(Double, r1, Reg(r3), fscratch_register))) block1;
               Dlist.insert_before (Opcode(F_Store(Double, fscratch_register, r2, Reg(r3)))) block1;
         );

         Dlist.insert_before (Opcode(Triadic(Sub, true, r3, Imm13(
           match typ with Word -> 4 | DoubleWord -> 8), r3))) block1;
         Dlist.paste_before (create_branch GreaterOrEqual label1) block1;
         Dlist.insert_before (Label(label2)) block1 );
         block1
;;

let map_temp_stack offset f_offset sequence =
  Dlist.p_iter (fun ins ->
    match Dlist.contents ins with
      Opcode(Load(Word, frame_pointer, Imm13(z), r)) when z>=announce_integer_stack ->
        let w=offset+(z-announce_integer_stack)*4
        in  Dlist.replace ins (Opcode(Load(Word, frame_pointer, Imm13(w), r)))
    | Opcode(Store(Word, r, frame_pointer, Imm13(z))) when z>=announce_integer_stack ->
        let w=offset+(z-announce_integer_stack)*4
        in  Dlist.replace ins (Opcode(Store(Word, r, frame_pointer, Imm13(w))))

    | Opcode(F_Load(Double, frame_pointer, Imm13(z), r)) when z>=announce_float_stack ->
        let w=f_offset+(z-announce_float_stack)*8
        in  Dlist.replace ins (Opcode(F_Load(Double, frame_pointer, Imm13(w), r)))
    | Opcode(F_Store(Double, r, frame_pointer, Imm13(z))) when z>=announce_float_stack ->
        let w=f_offset+(z-announce_float_stack)*8
        in  Dlist.replace ins (Opcode(F_Store(Double, r, frame_pointer, Imm13(w))))
    | _ -> () ) sequence;;

let generation arbre var_size identifier nb_args = 
  (*Printf.fprintf stderr "id = %s, var = %d, args = %d\n" identifier var_size nb_args;*)
  let procname = "proc__" ^ identifier
  and frame_label = "frame__" ^ identifier
  in
try
i_stack:=0; f_stack:=0;
let top_of_frame_size = 8 in
let test = match arbre with
    Iexpr (tree) -> int_eval tree
  | Fexpr (tree) -> float_eval tree
  | Vexpr (tree) -> void_eval tree
  | Bexpr (tree) -> int_eval (bool_to_int tree)
  | Pexpr (tree) -> ptr_eval tree
in
let i_stack_size = max 0 (!i_stack*4)
and f_stack_size = max 0 (!f_stack*8)
in
let i_stack_offset = top_of_frame_size+var_size+i_stack_size
in
let f_stack_offset = ((i_stack_offset + f_stack_size + 4)/8)*8
in let total_frame_size = f_stack_offset + bottom_of_frame_size
in 
( if total_frame_size > 4096
  then Error.codegen_error (Printf.sprintf
    "Stack frame too big (%d > 4096) [top=%d, var=%d, istack=%d, fstack=%d, bot=%d]"
    total_frame_size top_of_frame_size var_size i_stack_size f_stack_size bottom_of_frame_size ) );

let 
toto = Dlist.singleton (Opcode(Breakpoint)) in
Dlist.insert_before
  (Opcode(Triadic(Save, false, stack_pointer, Imm13(-total_frame_size), stack_pointer))) toto;

for i=0 to (min (pred nb_args) 5)
do
  Dlist.insert_before (Opcode(Store(Word, (input_registers+i), frame_pointer, Imm13(parameters_offset+4*i)))) toto
done;
(* TODO: fix that
Dlist.insert_before (Opcode(Sethi((Hi_l frame_label), scratch_register))) toto;
Dlist.insert_before (Opcode(Load(Word, scratch_register, (Lo_l frame_label), overflow_ret_register))) toto;
Dlist.insert_before (Opcode(Store(Word, overflow_ret_register, frame_pointer, (Imm13 save_old_frame_offset)))) toto;
Dlist.insert_before (Opcode(Store(Word, frame_pointer, scratch_register, (Lo_l frame_label)))) toto;
*)

Dlist.paste_before test toto;

(*
Dlist.insert_before (Opcode(Sethi((Hi_l frame_label), scratch_register))) toto;
Dlist.insert_before (Opcode(Load(Word, frame_pointer, (Imm13 save_old_frame_offset), overflow_ret_register))) toto;
Dlist.insert_before (Opcode(Store(Word, overflow_ret_register, scratch_register, (Lo_l frame_label)))) toto;

(
match arbre with
  Fexpr(_) ->
    if (rd <> 0) then 
    ( Dlist.insert_before (Opcode(F_Diadic(F_Move, rd, 0))) toto;
      Dlist.insert_before (Opcode(F_Diadic(F_Move, (succ rd), 1))) toto; ()
    );
    Dlist.insert_before (Opcode(Ret)) toto;
    Dlist.insert_before (Opcode(Triadic(Restore, false, 0, Reg(0), return_fregister))) toto

| _ ->
    Dlist.insert_before (Opcode(Ret)) toto;
    Dlist.insert_before (Opcode(Triadic(Restore, false, rd, Reg(0), return_register))) toto
);
*)

map_temp_stack (-i_stack_offset) (-f_stack_offset) toto;

Sparc.remove_duplicates toto;
Sparc.peephole toto;

print_string ("\t.local\t" ^ frame_label ^ "\n\t\t.common " ^ frame_label ^ ", 4, 4\n");

print_string (".section\t\".text\"\n\t.align 4\n\t.global\t" ^ procname
  ^ "\n\t.type\t" ^ procname ^ ", #function\n" ^ procname ^ " :\n" );
output_sequence stdout toto;
print_string (".Lend"^procname^" :\n\t.size " ^ procname ^ ", .Lend"^procname^" - " ^ procname ^ "\n")

with Failure(x) -> output_string stderr x; output_char stderr '\n';;

