(*
$Id: typeintermed.mli,v 1.1.1.1 1998/06/23 22:35:10 monniaux Exp $
*)


type iexpr = IAdd of iexpr*iexpr
           | ISub of iexpr*iexpr
           | IDiv of iexpr*iexpr
           | IMod of iexpr*iexpr
           | IMul of iexpr*iexpr
           | INeg of iexpr
           | IConst of int
           | I_peek of pexpr
           | Iprocedure of pexpr * expr list
           | I_ITE of bexpr*iexpr*iexpr
           | Iseq of vexpr*iexpr
           | IBlock of declaration*iexpr
           | Iinvseq of iexpr * vexpr
           | IVar of int

and fexpr  = FAdd of fexpr*fexpr
           | FSub of fexpr*fexpr
           | FDiv of fexpr*fexpr
           | FMul of fexpr*fexpr
           | FNeg of fexpr
           | FConst of float
           | F_from_i of iexpr
           | F_peek of pexpr
           | Fprocedure of pexpr * (expr list)
           | F_ITE of bexpr*fexpr*fexpr
           | Fseq of vexpr*fexpr
           | FBlock of declaration*fexpr
           | Finvseq of fexpr * vexpr
           | FVar of int

and pexpr  = PAdd of pexpr*iexpr
           | P_peek of pexpr
           | Pprocedure of pexpr * (expr list)
           | P_ITE of bexpr*pexpr*pexpr
           | Pseq of vexpr*pexpr
           | IAssignment of pexpr * iexpr
           | FAssignment of pexpr * fexpr
           | BAssignment of pexpr * bexpr
           | PAssignment of pexpr * pexpr
           | PBlock of declaration* pexpr
           | Var of int
           | Nil
           | FramePtr
           | Alloca of iexpr
           | AsmLabel of string
           | Pinvseq of pexpr * vexpr

and bexpr = 
             B_peek of pexpr
           | Bprocedure of pexpr * (expr list)
           | IInf of iexpr*iexpr
           | ISup of iexpr*iexpr
           | FInf of fexpr*fexpr
           | FSup of fexpr*fexpr
           | IEgal of iexpr*iexpr
           | BEgal of bexpr*bexpr
           | FEgal of fexpr*fexpr
           | PEgal of pexpr*pexpr
           | True
           | False
           | BAnd of bexpr*bexpr
           | BOr of bexpr*bexpr
           | BNot of bexpr
           | B_ITE of bexpr * bexpr * bexpr
           | Bseq of vexpr * bexpr
           | BBlock of declaration * bexpr
           | Binvseq of bexpr * vexpr
           | BVar of int

and vexpr =
             For of int * iexpr * iexpr * vexpr
           | While of bexpr * vexpr
           | Vprocedure of pexpr * (expr list)
           | V_from_i of iexpr
           | V_from_f of fexpr
           | V_from_b of bexpr
           | V_from_p of pexpr
           | Vseq of vexpr * vexpr
           | VBlock of declaration * vexpr
           | V_IT of bexpr * vexpr
           | V_ITE of bexpr * vexpr * vexpr
           | CopyBlock of pexpr * pexpr * iexpr (* offset des 4 derniers octets *)
           | FCopyBlock of pexpr * pexpr * iexpr (* offset des 8 derniers octets *)
           | Free
           | NNop
           | Erreur     (* Le code genere doit provoquer une erreur : affectation de tableaux *)
                              (* ayant des dimentions differentes *)
           | Errorexpr  (* attribue lors d'une erreur semantique pour poursuivre l'analyse *)

and expr = Iexpr of iexpr
           | Fexpr of fexpr
           | Bexpr of bexpr
           | Vexpr of vexpr
           | Pexpr of int * pexpr * result_type

and decl =
    VariableDecl of string * int * result_type
  | ArrayDecl of array_decl
  | IdentityDecl of string * int * expr * result_type
  | IdentityDeclex of string * int * expr * result_type
  | ProcedureDecl of procedure_decl

and declaration = decl list

and proc_content = External of string | Declared of expr

and procedure_decl =
  { proc_name : string ;
    proc_nb : int ;
    proc_args : (int * result_type) list ;
    proc_type : result_type ;
    proc_value: proc_content }

and result_type =
    Integer
  | Float
  | Boolean
  | Void
  | RefType of result_type
  | RefArrayType of result_type * int
  | ProcType of (result_type list) * result_type

and array_decl =
  { array_id : string;
     array_nb : int ;
    array_bounds : (iexpr * iexpr) list ;
    array_type : result_type
  }








